"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Builder = void 0;
const property_scalar_impl_1 = require("../../properties/property-scalar-impl");
const property_scalar_rule_builder_1 = require("./property-scalar-rule-builder");
const property_scalar_builder_1 = require("./property-scalar-builder");
const trigger_builder_1 = require("./trigger-builder");
const group_of_properties_builder_1 = require("./group-of-properties-builder");
const group_of_properties_impl_1 = require("../../properties/group-of-properties-impl");
const common_validators_1 = require("../../validators/common/common-validators");
const list_builder_1 = require("./list-builder");
const list_of_properties_impl_1 = require("../../properties/list-of-properties-impl");
const property_array_list_impl_1 = require("../../properties/property-array-list-impl");
const text_interpreter_1 = require("../../util/text-interpreter/text-interpreter");
class Builder {
    constructor(options, ruleEngine, dependencyGraph, propertyMap) {
        var _a;
        this.ruleEngine = ruleEngine;
        this.dependencyGraph = dependencyGraph;
        this.propertyMap = propertyMap;
        this.trigger = new trigger_builder_1.TriggerBuilder();
        this.notEmptyIfRequiredValidator = options.emptyButRequiredMessage instanceof Function
            ? common_validators_1.V.scalar.notEmptyMsgProvider(options.emptyButRequiredMessage)
            : common_validators_1.V.scalar.notEmpty(options.emptyButRequiredMessage);
        if (options.defaultEmptyChoiceDisplayValue) {
            this.defaultEmptyChoiceDisplayValue = options.defaultEmptyChoiceDisplayValue;
        }
        this.defaultBackpressureConfig = Object.freeze((_a = options.defaultBackpressureConfig) !== null && _a !== void 0 ? _a : {
            type: "switch",
            debounceTime: 40
        });
        this.textInterpreters = {
            [text_interpreter_1.TextInterpreter.Html]: options.textInterpreterHtml,
            [text_interpreter_1.TextInterpreter.Markdown]: options.textInterpreterMarkdown,
            [text_interpreter_1.TextInterpreter.Custom]: options.textInterpreterCustom,
        };
        this.list = new list_builder_1.ListBuilder((id, itemTemplate, config) => this.listOfProperties(id, itemTemplate, config), (id, provider, dependencies) => this.propertyList(id, provider, dependencies), (id, provider, dependencies, propertyConfig) => this.asyncPropertyList(id, provider, dependencies, propertyConfig));
        this.group = new group_of_properties_builder_1.GroupOfPropertiesBuilder((id, properties) => this.groupOfProperties(id, properties));
        this.scalar = new property_scalar_builder_1.PropertyScalarBuilder((id, provider, emptyValueFcn, converter, dependencies, propertyConfig, ownedProperties) => this.propertyScalar(id, provider, emptyValueFcn, converter, dependencies, propertyConfig, ownedProperties), (prop) => this.bindPropertyScalar(prop), this.defaultEmptyChoiceDisplayValue, this.list);
    }
    get properties() {
        return Object.values(this.propertyMap);
    }
    propertyScalar(id, provider, emptyValueFcn, converter, dependencies, config, ownedProperties) {
        var _a;
        const prop = new property_scalar_impl_1.PropertyScalarImpl(id, provider, emptyValueFcn, converter, this.ruleEngine, (_a = config === null || config === void 0 ? void 0 : config.backpressure) !== null && _a !== void 0 ? _a : (provider.isAsynchronous() ? this.defaultBackpressureConfig : undefined));
        this.addProperty(prop);
        if (dependencies) {
            this.addDependencies(this.dependencyGraph, dependencies, prop, { value: true });
        }
        if (ownedProperties) {
            ownedProperties.forEach(owned => this.dependencyGraph.addOwnerDependency(prop, owned, false));
        }
        return prop;
    }
    bindPropertyScalar(prop) {
        return new property_scalar_rule_builder_1.PropertyScalarRuleBuilder(prop, this.notEmptyIfRequiredValidator, (from, to, options) => this.addDependencies(this.dependencyGraph, from, to, options), this.textInterpreters);
    }
    groupOfProperties(id, properties) {
        const prop = new group_of_properties_impl_1.GroupOfPropertiesImpl(id, properties, this.ruleEngine);
        this.addProperty(prop);
        this.addDependencies(this.dependencyGraph, prop.propertiesAsList, prop, { value: true });
        return prop;
    }
    listOfProperties(id, itemTemplate, config) {
        const isMultiSelect = (config === null || config === void 0 ? void 0 : config.selectionMode) === list_builder_1.SelectionMode.MultiSelect;
        const prop = new list_of_properties_impl_1.ListOfPropertiesImpl(id, itemTemplate, isMultiSelect, this.ruleEngine, this.dependencyGraph);
        this.addProperty(prop);
        return prop;
    }
    propertyList(id, provider, dependencies) {
        const prop = new property_array_list_impl_1.PropertyArrayListSyncImpl(id, provider, this.ruleEngine, undefined);
        this.addProperty(prop);
        if (dependencies) {
            this.addDependencies(this.dependencyGraph, dependencies, prop, { value: true });
        }
        return prop;
    }
    asyncPropertyList(id, provider, dependencies, config) {
        const prop = new property_array_list_impl_1.PropertyArrayListAsyncImpl(id, provider, this.ruleEngine, config === null || config === void 0 ? void 0 : config.backpressure);
        this.addProperty(prop);
        if (dependencies) {
            this.addDependencies(this.dependencyGraph, dependencies, prop, { value: true });
        }
        return prop;
    }
    addDependencies(dependencyGraph, from, to, options) {
        if (from.length) {
            dependencyGraph.addDependencies(from, to, options);
        }
    }
    addProperty(property) {
        if (this.propertyMap[property.id]) {
            throw new Error(`A Property with the id ${property.id} is already defined`);
        }
        this.propertyMap[property.id] = property;
    }
    defineCustomAttribute(name) {
        return { name };
    }
    bindValidator(...properties) {
        return (validator) => {
            const instance = {
                getValidatedProperties: () => properties,
                validate: validator
            };
            properties.forEach((prop, i) => {
                prop.addValidator(instance);
                for (let j = i + 1; j < properties.length; j++) {
                    const dependent = properties[j];
                    this.dependencyGraph.addDependency(prop, dependent, { validation: true });
                    this.dependencyGraph.addDependency(dependent, prop, { validation: true });
                }
            });
        };
    }
    decorateTemplate(template, decorator) {
        return (id, index, siblingAccess) => {
            const property = template(id, index, siblingAccess);
            decorator(property);
            return property;
        };
    }
    asVisJsData() {
        return this.dependencyGraph.createVisJsData(this.properties);
    }
    generateNetworkGraphHtmlPage() {
        const graph = this.asVisJsData();
        const nodes = JSON.stringify(graph.nodes);
        const edges = JSON.stringify(graph.edges);
        const options = JSON.stringify(graph.options);
        return `<!DOCTYPE HTML>
<html>
<head>
    <script src="https://visjs.github.io/vis-network/standalone/umd/vis-network.min.js"></script>
    <style>
        #network {
            position: absolute; top: 0; right: 0; bottom: 0; left: 0;
        }
    </style>
</head>
<body>
    <div id="network"></div>
    <script type="text/javascript">
        var container = document.getElementById("network");
        var data = {
            nodes: new vis.DataSet(${nodes}),
            edges: new vis.DataSet(${edges}),
        };
        var options = ${options};
        var network = new vis.Network(container, data, options);
    </script> 
</body>
</html>`;
    }
}
exports.Builder = Builder;
