var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { DependencyGraph } from "../dependency-graph/dependency-graph";
import { Builder } from "./builder/builder";
import { ValidationType } from "../validators/validation-type";
import { RuleSet } from "./rule-set/rule-set";
export class RuleEngineImpl {
    constructor(options) {
        this.propertyMap = {};
        this.dependencyGraph = new DependencyGraph();
        this.validations = new WeakMap();
        this.dataLinks = new Map();
        this.snapshots = new Map();
        this.builder = new Builder(options, this, this.dependencyGraph, this.propertyMap);
    }
    get properties() {
        return Object.values(this.propertyMap);
    }
    getBuilder() {
        return this.builder;
    }
    getPropertyById(id) {
        return this.propertyMap[id];
    }
    validate() {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all(Object.values(this.propertyMap).map(prop => prop.validate()));
            return this.getValidationMessages();
        });
    }
    setValidationMessages(validationMessagesMap) {
        const unknownIds = [];
        this.clearValidationResult();
        Object.keys(validationMessagesMap).forEach(key => {
            if (this.propertyMap[key] !== undefined) {
                this.propertyMap[key].setValidationMessages(validationMessagesMap[key]);
            }
            else {
                unknownIds.push(key);
            }
        });
        return unknownIds;
    }
    getValidationMessages() {
        return Object.keys(this.propertyMap).reduce((res, cur) => {
            const messages = this.propertyMap[cur].getValidationMessages();
            if (messages.length > 0) {
                res[cur] = messages;
            }
            return res;
        }, {});
    }
    clearValidationResult() {
        Object.values(this.propertyMap).forEach(prop => prop.clearValidationResult());
    }
    takeSnapShot(key = 'default') {
        const snap = {
            key,
            data: this.properties.map(p => p.exportData())
        };
        this.snapshots.set(key, snap);
        return snap;
    }
    restoreSnapShot(key = 'default') {
        const snap = this.snapshots.get(key);
        if (snap) {
            snap.data.forEach((d, i) => this.properties[i].importData(d));
        }
    }
    linkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        if (this.dataLinks.has(dataLinkKey)) {
            return;
        }
        const listenerA = {
            updated: () => propertyA.transferData(propertyB)
        };
        const listenerB = {
            updated: () => propertyB.transferData(propertyA)
        };
        propertyA.transferData(propertyB);
        const refA = propertyA.registerValueChangedListener(listenerA);
        const refB = propertyB.registerValueChangedListener(listenerB);
        this.dataLinks.set(dataLinkKey, [refA, refB]);
    }
    unlinkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        const listeners = this.dataLinks.get(dataLinkKey);
        if (listeners) {
            const [refA, refB] = listeners;
            propertyA.deregisterValueChangedListener(refA);
            propertyB.deregisterValueChangedListener(refB);
            this.dataLinks.delete(dataLinkKey);
        }
    }
    createDataLinkKey(propertyA, propertyB) {
        return propertyA.id < propertyB.id
            ? `${propertyA.id}<->${propertyB.id}`
            : `${propertyB.id}<->${propertyA.id}`;
    }
    needsAnUpdate(mightHaveChanged) {
        this.dependencyGraph.traverseDepthFirst(mightHaveChanged.id, property => {
            property.needsAnUpdate(false);
        }, dependency => !!dependency.options.value);
    }
    updateValue(property) {
        var _a;
        const asyncUpdates = (_a = this.dependencyGraph.getAsyncDependencies(property.id)) === null || _a === void 0 ? void 0 : _a.map(asyncDep => this.updateValue(asyncDep));
        if ((asyncUpdates === null || asyncUpdates === void 0 ? void 0 : asyncUpdates.length) && asyncUpdates.some(update => !!update)) {
            return Promise.all(asyncUpdates).then(() => this.updateValueInternal(property)).catch((err) => {
                property.errorWhileUpdating(err);
            });
        }
        else {
            return this.updateValueInternal(property);
        }
    }
    updateValueInternal(property) {
        try {
            const updatePromise = property.internallyUpdate();
            if (updatePromise) {
                return updatePromise.then(() => {
                    this.hasBeenUpdated(property);
                }).catch(err => property.errorWhileUpdating(err));
            }
            else {
                this.hasBeenUpdated(property);
            }
        }
        catch (err) {
            property.errorWhileUpdating(err);
        }
    }
    hasBeenUpdated(property) {
        property.hasBeenUpdated();
        this.dependencyGraph.traverseDepthFirst(property.id, (prop, dep) => prop.dependencyHasBeenUpdated(dep), dependency => !dependency.options.value, true);
    }
    cancelValidationAndInvalidateResults(validators) {
        validators.forEach(v => {
            const vprocess = this.getValidationProcess(v);
            vprocess.isLastResultUpToDate = false;
            vprocess.isCancelled = true;
        });
    }
    validateValidatorInstances(validators) {
        return validators.map(validator => {
            const vprocess = this.getValidationProcess(validator);
            if (vprocess.isLastResultUpToDate) {
                return Promise.resolve(vprocess.lastValidationResult);
            }
            else if (vprocess.currentValidation && !vprocess.isCancelled) {
                return vprocess.currentValidation;
            }
            vprocess.isCancelled = false;
            const validation = validator.validate(...validator.getValidatedProperties());
            if (validation instanceof Promise) {
                vprocess.currentValidation = validation.then(res => {
                    vprocess.currentValidation = undefined;
                    vprocess.lastValidationResult = res;
                    if (vprocess.isCancelled) {
                        return 'cancelled';
                    }
                    else {
                        vprocess.isLastResultUpToDate = true;
                        return res;
                    }
                }, err => {
                    vprocess.currentValidation = undefined;
                    return [{
                            text: JSON.stringify(err),
                            type: ValidationType.ErrorThrownDuringValidation
                        }];
                });
                return vprocess.currentValidation;
            }
            else {
                vprocess.lastValidationResult = validation;
                vprocess.isLastResultUpToDate = true;
                return Promise.resolve(validation);
            }
        });
    }
    getValidationProcess(validator) {
        const vprocess = this.validations.get(validator);
        if (vprocess) {
            return vprocess;
        }
        else {
            const newProcess = {
                isLastResultUpToDate: false,
                isCancelled: false,
            };
            this.validations.set(validator, newProcess);
            return newProcess;
        }
    }
    defineRuleSet(initFcn) {
        return new RuleSet(initFcn, this.builder);
    }
}
